/*
 *	netcam.c
 *
 *	Copyright 2002 by Jeroen Vreeken (pe1rxq@amsat.org)
 *	This software is distributed under the GNU Public License Version 2
 *	See also the file 'COPYING'.
 *
 */

#include "motion.h"

#ifdef __freebsd__
#include "video_freebsd.h"
#else
#include "video.h"
#endif /* __freebsd__ */

#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <jpeglib.h>
#include <setjmp.h>

/* for rotation */
#include "rotate.h"
#include "netcam_wget.h"

#define REQUEST_AUTH "GET %s HTTP/1.0\r\nAuthorization: Basic %s\r\n\r\n"
#define REQUEST "GET %s HTTP/1.0\r\n\r\n"
#define TIMEOUT 30

typedef struct {
	struct jpeg_source_mgr pub;
	char *data;
	int length;
	JOCTET *buffer;
	boolean start_of_file;
} my_source_mgr;

typedef my_source_mgr *my_src_ptr;

struct my_error_mgr {
	struct jpeg_error_mgr pub;
	jmp_buf setjmp_buffer;
};
 
void my_init_source (j_decompress_ptr); 
boolean my_fill_input_buffer (j_decompress_ptr); 
void my_skip_input_data (j_decompress_ptr, long);
void my_term_source (j_decompress_ptr);
void jpeg_memory_src(j_decompress_ptr, char *, int);
void my_error_exit (j_common_ptr);

static void *netcam_loop(void *);
int netcam_read_header(struct netcam_context *, int);
int netcam_connect(struct netcam_context *);
int netcam_close(struct netcam_context *);
int netcam_read_image(struct netcam_context *);
int netcam_single_read(struct netcam_context *,int);
int netcam_read_image_no_content_length(struct netcam_context *);
int netcam_read_stream_no_content_length(struct netcam_context *);
int netcam_stream_read(struct netcam_context *,int);
	
int netcam_start(struct context *);
unsigned char *netcam_next(struct context *, char *);

/*
	parses the url
*/

int netcam_parse_url (struct url_t *parse_url, char *text_url)
{
	char *dup;
	char *parse;
	char *p;

	dup = strdup(text_url);
	parse = dup;
	parse_url->service = NULL;
	parse_url->host = NULL;
	parse_url->port = 80;
	parse_url->path = NULL;

	if ((p = strstr (parse, ":/"))) {
		*p++ = '\0';
		if (p[0] == '/' && p[1] == '/') p += 2;
		parse_url->service = strdup (dup);
		parse = p;
	}

	p = strchr (parse, '/');

	if (!p || parse < p){
		char *p2;

		parse_url->host = strdup (parse);
		if (p) parse_url->host[p - parse] = '\0';
		if (*parse_url->host == '[') {
			p2 = strchr (parse_url->host, ']');
			if (p2)	p2 = strchr (p2, ':');
		} else {
			p2 = strchr (parse_url->host, ':');
		}
		if (p2) {
			*p2++ = '\0';
			parse_url->port = atoi (p2);
		}
	}
	parse = p;
	if (parse) parse_url->path = strdup (parse);

	free (dup);

	return 1;
}


/* 
	jpeglib init_source 

	overrided by own init_source routine.
	Initialize source --- called by jpeg_read_header
	before any data is actually read.
*/

void my_init_source (j_decompress_ptr cinfo) 
{
	my_src_ptr src = (my_src_ptr) cinfo->src;
	/* We reset the empty-input-file flag for each image,
	but we don't clear the input buffer.
	This is correct behavior for reading a series of images from one source.
	*/
	src->start_of_file = TRUE;
}

/*
	jpeglib fill_input_buffer

	override by own fill_input_buffer routine.
	Fill the input buffer --- called whenever buffer is emptied
*/

boolean my_fill_input_buffer (j_decompress_ptr cinfo) 
{
	my_src_ptr src = (my_src_ptr) cinfo->src;
	size_t nbytes;

	if (src->start_of_file) {
		nbytes=src->length;
		src->buffer=src->data;
	} else {
	// Insert a fake EOI marker - as per jpeglib recommendation
		src->buffer[0] = (JOCTET) 0xFF;
		src->buffer[1] = (JOCTET) JPEG_EOI;
		nbytes = 2;
	}

	src->pub.next_input_byte = src->buffer;
	src->pub.bytes_in_buffer = nbytes;
	src->start_of_file = FALSE;
	
	return TRUE;
}

/*
	jpeglib skip_input_data 

	override by own skip_input_data routine. 
	Skip data --- used to skip over a potentially large amount of
	uninteresting data
*/

void my_skip_input_data (j_decompress_ptr cinfo, long num_bytes) 
{
my_src_ptr src = (my_src_ptr) cinfo->src;

	if (num_bytes > 0) {
		while (num_bytes > (long) src->pub.bytes_in_buffer) {
			num_bytes -= (long) src->pub.bytes_in_buffer;
			(void) my_fill_input_buffer(cinfo);
		}
	src->pub.next_input_byte += (size_t) num_bytes;
	src->pub.bytes_in_buffer -= (size_t) num_bytes;
	}
}

/* 
	jpeglib term_source 

	Terminate source --- called by jpeg_finish_decompress
	after all data has been read.  Often a no-op.
	NOT called by jpeg_abort or jpeg_destroy; surrounding
	application must deal with any cleanup that should happen even
	for error exit.
*/
void my_term_source (j_decompress_ptr cinfo) { }

/* 
	jpeg_memory_src 

	Prepare for input from memory .
	The caller must have already allocated memory, and is responsible
	for freed it after finishing decompression.
*/

void jpeg_memory_src(j_decompress_ptr cinfo, char *data, int length) 
{
	my_src_ptr src;

	if (cinfo->src == NULL) {
		cinfo->src = (struct jpeg_source_mgr *) (*cinfo->mem->alloc_small) 
			((j_common_ptr) cinfo, JPOOL_PERMANENT,sizeof(my_source_mgr));
		src = (my_src_ptr) cinfo->src;
	}

	src = (my_src_ptr) cinfo->src;
	src->data = data;
	src->length = length;
	src->pub.init_source = my_init_source;
	src->pub.fill_input_buffer = my_fill_input_buffer;
	src->pub.skip_input_data = my_skip_input_data;
	src->pub.resync_to_restart = jpeg_resync_to_restart;
	src->pub.term_source = my_term_source;
	src->pub.bytes_in_buffer = 0;
	src->pub.next_input_byte = NULL;
}


/* 
	jpeglib error_exit

	overrided by our own error_exit routine. 
	Triggered when a fatal error occurs.
*/

void my_error_exit (j_common_ptr cinfo)
{
	struct my_error_mgr *myerr = (struct my_error_mgr *) cinfo->err;
	(*cinfo->err->output_message) (cinfo);
	/* Return control to the setjmp point */
	longjmp (myerr->setjmp_buffer, 1);
}

/*
	main thread   
*/

static void *netcam_loop(void *arg)
{
	struct context *cnt = arg;
	syslog(LOG_DEBUG, "Netcam: thread %d: thread PID: %d",cnt->threadnr, getpid());
	printf("Netcam: thread %d: thread PID: %d\n", cnt->threadnr, getpid());
	do{
		cnt->netcam->read(cnt->netcam,0);
	}while (!cnt->finish);
	return 0;
}

/*
	if Content-Length is not supplied i.ex Axis 2100 , this function have to get 
	the boundary string for mjpeg.
	i.ex "Content-Type: multipart/x-mixed-replace; boundary=--myboundary".
*/

int netcam_read_header(struct netcam_context *netcam, int initial)
{
	char *header,*conttype;
	int status,hcount=0;
	
	while (1) {
		status = header_get (netcam->response, &header, HG_NONE);
		
/*
		// Removed , seems that doesn't working at all
		if (status == HG_ERROR || status == HG_EOF) {
			syslog(LOG_ERR, "Netcam: error reading header\nheader:\n [%s]\n",header);
			printf ("Netcam: error reading header\nheader:\n [%s]\n",header);
			free(header);
			return -1;
		}
*/	
		if (hcount == 0 && initial == 1) {
			/* process http ok ... or any other error header response */
			printf("[%s]\n",header);
		}
		
		if ( (header_process (header, "Content-Length", header_extract_number, 
			&netcam->image[netcam->which_image].size) ) ){
//			printf("Header content_length found size [%i]\n",netcam->image[netcam->which_image].size);	
			netcam->content_length=1;	
		}else if ((initial == 1) && (header_process (header, "Content-Type",
			http_process_type, &conttype) ))
		{
			if (strcmp("multipart/x-mixed-replace", conttype) == 0) {
				char *bound;
				fprintf(stderr, "Netcam: mjpeg stream\n");
				/* Set the bondary string following Content-Type */
				if ( (!netcam->boundary_string) && (bound = strstr(header,"boundary=")) ) {
					bound = bound + 9;
					netcam->boundary_string = strdup(bound);
					fprintf(stderr, "Netcam: netcam->boundary_string [%s]\n",netcam->boundary_string);
				}	
				netcam->read = (void *) netcam_stream_read;
			} else if (strcmp("image/jpeg", conttype) == 0) {
				fprintf(stderr, "Netcam: jpeg file\n");
				netcam->read = (void *) netcam_single_read;
			} else {
				syslog(LOG_ERR, "Netcam: error Uknown Content-Type");
				printf ("Netcam: error Uknown Content-Type\n");
				free(header);
				return -1;
			}
			if (conttype) free(conttype);
		}else if (!*header) {
			free (header);
			break;
		}
										
	
		hcount++;
		free (header);
	}
			
	return hcount;
}

/*
    Establish netcam connection , it has a 5 seconds timeout
*/

int netcam_connect(struct netcam_context *netcam)
{
	int sock;
	struct sockaddr_in server;
	struct hostent *host_info;
	struct timeval tv;	
	char *request;

	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		syslog(LOG_ERR, "Netcam: failed to create socket");
		printf("Netcam: failed to create socket\n");
		return -1;
	}

	memset(&server, 0, sizeof(server));
	host_info = gethostbyname(netcam->url.host);
	if (host_info == NULL) {
		syslog(LOG_ERR, "Netcam: unknown server: %s", netcam->url.host);
		printf("Netcam: unknown server: %s\n", netcam->url.host);
		return -1;
	}
	
	memcpy((char *) &server.sin_addr, host_info->h_addr, host_info->h_length);

	server.sin_family = AF_INET;
	server.sin_port = htons(netcam->url.port);

	/* 5 seconds timeout */
	tv.tv_sec = 5;
	tv.tv_usec = 0;
	setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv));
	
	if (connect (sock, (struct sockaddr *) &server, sizeof(server)) < 0) {
		syslog(LOG_ERR, "Netcam: can't connect to server");
		printf("Netcam: can't connect to server\n");
		return -1;
	}

	if (netcam->userpass){
		request = (char *) mymalloc(strlen(REQUEST_AUTH) +
			strlen(netcam->url.path) + strlen(netcam->userpass) - 4 + 1);
		sprintf(request, REQUEST_AUTH, netcam->url.path, netcam->userpass);
	}else{
		request = (char *) mymalloc(strlen(REQUEST) +
			strlen(netcam->url.path) + 1);
		sprintf(request, REQUEST, netcam->url.path);	
	}

	rbuf_initialize(netcam->response, sock);

	send(RBUF_FD(netcam->response), request, strlen(request), 0);

	free(request);
	
	return 0;
}

/*
   Close netcam connection , only for single image.
*/
int netcam_close(struct netcam_context *netcam)
{
	close(RBUF_FD(netcam->response));
	return 0;
}

/*
    AXIS 2100 
    --myboundary^M
    Content-Type: image/jpeg^M
    <image>^M
    --myboundary^M
    Content-Type: image/jpeg^M
    <image>^M
    [ .... ]
 
	netcam stream read function for Non Content-Length support devices.
	On Success returns 1
	On Error returns -1
*/
int netcam_read_stream_no_content_length(struct netcam_context *netcam)
{
	int which_image,cur_size=0,readb=0;
	size_t boundary_string_length;
	char *pointer_boundary=NULL;

	which_image = netcam->which_image;
	
	/* FIXME !? */
	cur_size = rbuf_flush(netcam->response, netcam->image[which_image].buffer,50000);

	boundary_string_length = strlen(netcam->boundary_string); 

	/* 
		Get the image until find the next STREAM HEADER 
	*/

	do {
		readb = read(netcam->response->fd, &netcam->image[which_image].buffer[cur_size], 4096);
		if (readb < 0){
			return -1;
		}
		cur_size += readb;
		if ( (pointer_boundary = memmem(netcam->image[which_image].buffer,cur_size,netcam->boundary_string,
		                                boundary_string_length)) ) {
			cur_size = pointer_boundary - netcam->image[which_image].buffer;
		}
	} while (!pointer_boundary);

	netcam->image[which_image].buffer_size = cur_size -4;
	netcam->image[which_image].size = cur_size -4;
		
	return 1; 
}


/*
	netcam image read function for Non Content-Length support devices
	On Success returns 0
	On Error returns -1
*/
int netcam_read_image_no_content_length(struct netcam_context *netcam)
{
	int which_image,cur_size,readb;
		
	which_image = netcam->which_image;
	
	/* FIXME !? */
	cur_size = rbuf_flush(netcam->response, netcam->image[which_image].buffer, 50000);

	do{	
		readb = read(netcam->response->fd, &netcam->image[which_image].buffer[cur_size], 4096);
		cur_size += readb;
	} while (readb > 0);

	if (readb < 0){
		return -1;
	}
	netcam->image[which_image].buffer_size = cur_size;
	netcam->image[which_image].size = cur_size;
	
	return 0;
}


/*
	netcam stream/image read function for Content-Length support devices

*/
int netcam_read_image(struct netcam_context *netcam)
{
	int which_image,cur_size,cur_left;
	
	which_image = netcam->which_image;
	if (netcam->image[which_image].size > netcam->image[which_image].buffer_size){
		netcam->image[which_image].buffer_size = netcam->image[which_image].size;
		netcam->image[which_image].buffer = myrealloc(netcam->image[which_image].buffer,
				netcam->image[which_image].buffer_size,"netcam_stream_read");
	}

	cur_size = rbuf_flush(netcam->response, netcam->image[which_image].buffer,
							netcam->image[which_image].buffer_size);

	while (netcam->image[which_image].size > cur_size) {
		cur_left = recv(netcam->response->fd,&netcam->image[which_image].buffer[cur_size],
						netcam->image[which_image].size - cur_size, 0);
		if (cur_left < 0){
			syslog(LOG_ERR, "Netcam: error reading image, not complete");
			printf ("Netcam: error reading image, not complete\n");
			netcam->image[which_image].size = 0;
			return -1;
		}
		cur_size += cur_left;		
	}
	
	return 0;
}	

/*
	netcam image read function 
*/
int netcam_single_read(struct netcam_context *netcam,int initial)
{
	int which_image,ret;
	
	if (netcam_connect(netcam) != 0) return -1;
	pthread_mutex_lock(&netcam->mutex);
	which_image = netcam->which_image = !netcam->which_image;
	pthread_mutex_unlock(&netcam->mutex);
	do {
		if (initial) ret = netcam_read_header(netcam, 2);
		else{	
			ret = netcam_read_header(netcam, 0);
		}	
	} while (ret == 0);
	pthread_mutex_lock(&netcam->image[which_image].mutex);
	if (netcam->content_length) netcam_read_image(netcam);
	else netcam_read_image_no_content_length(netcam); 
	pthread_mutex_unlock(&netcam->image[which_image].mutex);	
	netcam_close(netcam);
	
	return 0;
}

/*
	netcam stream read function 
*/
int netcam_stream_read(struct netcam_context *netcam,int initial)
{
	int which_image;
	int ret;

	pthread_mutex_lock(&netcam->mutex);
	which_image = netcam->which_image = !netcam->which_image;
	pthread_mutex_unlock(&netcam->mutex);
	do {
		if (initial) ret = netcam_read_header(netcam, 0);
		else {
			if (netcam->content_length) ret = netcam_read_header(netcam, 0);
			else ret = netcam_read_header(netcam, 2);
		}	
	} while (ret == 0);
	pthread_mutex_lock(&netcam->image[which_image].mutex);
	if (netcam->content_length) netcam_read_image(netcam);
	else netcam_read_stream_no_content_length(netcam);
	pthread_mutex_unlock(&netcam->image[which_image].mutex);
	
	return 0;
}


/*
	netcam init called function , it establishes connection to network device, gets
	the image or the stream and creates the read thread.
*/

int netcam_start(struct context *cnt)
{
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	pthread_attr_t thread_attr;
	pthread_t thread_id;
	struct netcam_context *netcam;

	syslog(LOG_INFO, "Netcam: thread %d: starting...", cnt->threadnr);
	printf("Netcam: thread %d: starting...\n", cnt->threadnr);

	cnt->netcam = (struct netcam_context *) mymalloc(sizeof(struct netcam_context));
	netcam = cnt->netcam;

	pthread_mutex_init(&netcam->mutex, NULL);
	pthread_mutex_init(&netcam->image[0].mutex, NULL);
	pthread_mutex_init(&netcam->image[1].mutex, NULL);

	netcam_parse_url (&netcam->url, cnt->conf.netcam_url);

	if (cnt->conf.netcam_userpass){
		netcam->userpass = (char *) mymalloc(BASE64_LENGTH(strlen(cnt->conf.netcam_userpass))+1);
		{
			char *userpass=mymalloc(strlen(cnt->conf.netcam_userpass)+4);
			/* base64_encode can read 3 bytes after the end of the string,
			initialize it */
			memset(userpass,0,strlen(cnt->conf.netcam_userpass)+4);
			strcpy(userpass,cnt->conf.netcam_userpass);
			base64_encode(userpass, netcam->userpass, strlen(cnt->conf.netcam_userpass));
			free(userpass);
		}
	}	
	
	netcam->response = (struct rbuf *) mymalloc(sizeof(struct rbuf));

	/* 
		image buffer should be set for Non Content-Length support devices.
		FIXME : Maybe it should not be hardcoded here !?
	*/
	
	netcam->image[0].buffer_size = 50000;
	netcam->image[0].size = 0;
	netcam->image[0].buffer = (char *) mymalloc(netcam->image[0].buffer_size);
	netcam->image[1].buffer_size = 50000;
	netcam->image[1].size = 0;
	netcam->image[1].buffer = (char *) mymalloc(netcam->image[1].buffer_size);
	netcam->which_image = 0;

	/* Initial Connection to Netcam */
	if (netcam_connect(netcam) != 0) return -1;
	
	/* Initial header read , as specified passing as parameter 1 */	
	if ( netcam_read_header(netcam, 1) == -1){
		return -1;
	}
	
	/* Call to fetch image or stream function */
	netcam->read(netcam,1);

	if (netcam->image[netcam->which_image].size > 0) {

		cinfo.err = jpeg_std_error(&jerr);
		jpeg_create_decompress(&cinfo);
		jpeg_memory_src(&cinfo,	netcam->image[netcam->which_image].buffer,
						netcam->image[netcam->which_image].size);
		jpeg_read_header(&cinfo, TRUE);

		/* NOTE: The conf values shouldn't be set here, since then they won't reflect
			the configuration file. Changed from 3.1.18-snap2.

			cnt->conf.width=cinfo.image_width;
			cnt->conf.height=cinfo.image_height;
		*/

		cnt->imgs.width=cinfo.image_width;
		cnt->imgs.height=cinfo.image_height;
		cnt->imgs.size=cnt->imgs.width*cnt->imgs.height*3/2;
		cnt->imgs.motionsize=cnt->imgs.width*cnt->imgs.height;
		cnt->imgs.type=VIDEO_PALETTE_YUV420P;

		/* Create the netcam loop thread */
		pthread_attr_init(&thread_attr);
		pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);
		pthread_create(&thread_id, &thread_attr, &netcam_loop, cnt);
	}else{
		printf("Image size %d\n",netcam->image[netcam->which_image].size);
		return -1;	
	}
	return 0;
}


/*
	netcam process image function , is called by motion thread to get an image
	from netcam read thread.
*/

unsigned char *netcam_next(struct context *cnt, char *image)
{
	struct jpeg_decompress_struct cinfo;
	struct my_error_mgr jerr;
	unsigned char *pic, *upic, *vpic;
	JSAMPARRAY line;
	int i, line_size, y,which_image;
	JSAMPROW row[1];
	int width, height;		/* for rotation */
	struct netcam_context *netcam;
	
	pic = NULL;
	netcam = cnt->netcam;
	pthread_mutex_lock(&netcam->mutex);
	which_image = !netcam->which_image;
	pthread_mutex_unlock(&netcam->mutex);
	pthread_mutex_lock(&netcam->image[which_image].mutex);
	
	if (netcam->image[which_image].size > 0) {
		cinfo.err = jpeg_std_error(&jerr.pub);
		jerr.pub.error_exit = my_error_exit;

		if (setjmp(jerr.setjmp_buffer)) {
			jpeg_destroy_decompress(&cinfo);
			pthread_mutex_unlock(&netcam->image[which_image].mutex);
			return pic;
		}

		jpeg_create_decompress(&cinfo);
		jpeg_memory_src(&cinfo, netcam->image[which_image].buffer,netcam->image[which_image].size);
		jpeg_read_header(&cinfo, TRUE);

		cinfo.out_color_space = JCS_YCbCr;

		/* NOTE: Since this is a capture, we need to use the capture dimensions. */
		/* POSSIBLE BUG: What if the netcam changes resolution? */

		width = cnt->rotate_data.cap_width;
		height = cnt->rotate_data.cap_height;
	
		jpeg_start_decompress(&cinfo);
		line_size = width * 3;

		line = (*cinfo.mem->alloc_sarray) ((j_common_ptr) & cinfo,
			JPOOL_IMAGE, cinfo.output_width * cinfo.output_components, 1);
		pic = image;
		upic = pic + width * height;
		vpic = upic + (width * height) / 4;
		row[0] = (unsigned char *) line;
		y = 0;
		
		while (cinfo.output_scanline < height) {
			jpeg_read_scanlines(&cinfo, row, 1);
			for (i = 0; i < line_size; i += 3) {
				pic[i / 3] = ((unsigned char *) line)[i];
				if (i & 1) {
					upic[(i / 3) / 2] = ((unsigned char *) line)[i + 1];
					vpic[(i / 3) / 2] = ((unsigned char *) line)[i + 2];
				}
			}
			pic += line_size / 3;
			if (y++ & 1) {
				upic += width / 2;
				vpic += width / 2;
			}
		}
		
		jpeg_finish_decompress(&cinfo);
		jpeg_destroy_decompress(&cinfo);

		if(cnt->rotate_data.degrees > 0) {
			/* rotate as specified */
		rotate_map(image, cnt);
		}
	}
	pthread_mutex_unlock(&netcam->image[which_image].mutex);

	return pic;
}
